/****************************************************************************
**  SIONLIB     http://www.fz-juelich.de/jsc/sionlib                       **
*****************************************************************************
**  Copyright (c) 2008-2019                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/
#define _XOPEN_SOURCE 700

#include "sion_hints_linux.h"
#include "sion_const.h"
#include "sion_debug.h"

int _sion_apply_hint_linux_access_range(int fd, long long start, long long length)
{
  int rc = SION_SUCCESS;


  return rc;
}
