/****************************************************************************
**  SIONLIB     http://www.fz-juelich.de/jsc/sionlib                       **
*****************************************************************************
**  Copyright (c) 2008-2019                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

/// \cond sion_internal
/// \file

#ifndef SION_SION_HINTS_LINUX_H
#define SION_SION_HINTS_LINUX_H

/*!\brief allocates and initalize the hints
 *
 * @param  fd      file descriptor
 * @param  start   start
 * @param  length  length
 *
 * @return SION_SUCCESS if ok
 */
int _sion_apply_hint_linux_access_range(int fd, long long start, long long length);

#endif

/// \endcond
