/****************************************************************************
**  SIONLIB     http://www.fz-juelich.de/jsc/sionlib                       **
*****************************************************************************
**  Copyright (c) 2008-2019                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

/*!
 * \cond sion_internal
 * \file
 *
 */

#ifndef SION_SION_FD_H
#define SION_SION_FD_H

#include "sion_filedesc.h"

#define SION_FILEDESCRIPTOR 11 /*!< int id -> data structure mapping (from visit library) */
#define SION_APIDESCRIPTOR 12 /*!< int id -> data structure mapping (from visit library) */

#ifdef __cplusplus
extern "C" {
#endif

int _sion_newvcd(void *data, int type);
int _sion_freevcd(int sid);
void *_sion_vcdtovcon(int sid);
int _sion_vcdtype(int sid);
int _sion_reassignvcd(int sid, void *data, int type);

_sion_filedesc *_sion_get_filedesc(int sid);

#ifdef __cplusplus
}
#endif

#endif

/// \endcond
