/****************************************************************************
**  SIONLIB     http://www.fz-juelich.de/jsc/sionlib                       **
*****************************************************************************
**  Copyright (c) 2008-2019                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

/*!
 * \cond sion_internal
 * \file
 */

#ifndef SION_SION_OMP_INTERNAL_GEN_H
#define SION_SION_OMP_INTERNAL_GEN_H

#ifdef SION_OMP

#include "sion_generic.h"
#include "sion_generic_internal.h"

int _sion_errorprint_omp(int rc, int level, const char *format, ...);

struct sion_omp_options {
  sion_generic_options generic_options;
};

#define SION_OMP_OPTIONS_INIT \
  (sion_omp_options) \
  { \
    .generic_options = SION_GENERIC_OPTIONS_INIT, \
  }

#endif

#endif

/// \endcond
